!
! Sample program for  Kathmandu_interface_grid.csv
!                   by K. Koketsu (Nov. 10, 2023) 
! <Data Format>
! UTM_x,UTM_y,Longitude,Lattitude,Heights(from the ground) of Interfaces 1,2,...,7,
! and Height(from the sea level) of the ground  

  implicit none
  real(8):: phi, lam, pi, xpp, ypp, ground
  real(8):: xx0, yy0, dxx, dyy, xx, yy, lon, lat
  integer(4):: jx, jy, ix, nx, iy, ny, nhi, i
  parameter (dxx=200.d0, dyy=200.d0)
  parameter (nx=353, xx0= 304000.d0) ! x = 304000, 304200, ...., 374400 (xx0+dxx*(nx-1)) 
  parameter (ny=306, yy0=3034800.d0) ! y =3034800,3035000, ....,3095800 (yy0+dyy*(ny-1))
  parameter (nhi=7)
  real(8):: h(nhi), h11(nhi), h12(nhi), h21(nhi), h22(nhi)
  real(8):: xpp1, xpp2, ypp1, ypp2

!  Initialization
  pi = acos(-1.0d0)

!  Target point
!  example: the center of Kathmandu Valley  
  lam = 85.3d0
  phi = 27.7d0
  write(6,'(2f10.4)') lam, phi
!  converted by http://rcn.montana.edu/resources/Converter.aspx
  xpp =  332377.d0
  ypp = 3065127.d0
  write(6,'(2f15.5)') xpp, ypp
  jx = int((xpp - xx0) / dxx) + 1
  jy = int((ypp - yy0) / dyy) + 1
  write(6,'(2i15)') jx, jy 

!  Reading of the model
  open(21,file="Kathmandu_interface_grid.csv")
  do iy=1,jy+1
     do ix=1,nx
        read(21,*) xx, yy, lon, lat, (h(i),i=1,nhi), ground
        if(ix.eq.jx .and. iy.eq.jy) then
           xpp1 = xx
           ypp1 = yy
           do i=1,nhi
              h11(i) = h(i) + ground
           enddo
           write(6,'(2i5,2f15.0,2f15.5)') ix, iy, xx, yy, lon, lat
        else if(ix.eq.jx+1 .and. iy.eq.jy) then
           do i=1,nhi
              h21(i) = h(i) + ground
           enddo
           write(6,'(2i5,2f15.0,2f15.5)') ix, iy, xx, yy, lon, lat
        else if(ix.eq.jx .and. iy.eq.jy+1) then
           do i=1,nhi
              h12(i) = h(i) + ground
           enddo
           write(6,'(2i5,2f15.0,2f15.5)') ix, iy, xx, yy, lon, lat
        else if(ix.eq.jx+1 .and. iy.eq.jy+1) then
           xpp2 = xx
           ypp2 = yy
           do i=1,nhi
              h22(i) = h(i) + ground
           enddo
           write(6,'(2i5,2f15.0,2f15.5)') ix, iy, xx, yy, lon, lat
         endif
      enddo
  enddo

! Linear interpolation
      do i=1,nhi
        h(i) = h11(i)*(xpp2-xpp)/(xpp2-xpp1)*(ypp2-ypp)/(ypp2-ypp1) &
           & + h21(i)*(xpp-xpp1)/(xpp2-xpp1)*(ypp2-ypp)/(ypp2-ypp1) &
           & + h12(i)*(xpp2-xpp)/(xpp2-xpp1)*(ypp-ypp1)/(ypp2-ypp1) &
           & + h22(i)*(xpp-xpp1)/(xpp2-xpp1)*(ypp-ypp1)/(ypp2-ypp1)
        write(6,'(i2,a,f13.5,a,4f13.5)') i,'|',h(i),' <--',h11(i),h21(i),h12(i),h22(i)
      enddo

end program

subroutine bl2xy(phi,lam,KEI,xx,yy)
! written by Hiroaki Kobayashi (Kobori Research Complex) using the GSI formulation 
  implicit none
  integer(4)::KEI
  real(8)::phi,lam,xx,yy
!  phi:lattitude (radian)
!  lam:longitude (rarian)
!  xx : north is positive (m)
!  yy : east is positive (m)
  integer(4)::ibd(20),ld(20),lm(20)
  real(8)::pi,a,f,m0,n,phi0,lam0,temp,t,tvar
  real(8)::lamc,lams,gzd,etad,alpha1,alpha2,alpha3,alpha4,alpha5
  real(8)::a0,a1,a2,a3,a4,a5,sphi0var,avar
  data ibd/33,33,36,33,36,36,36,36,36,40,44,44,44,26,26,26,26,20,26,28/
  data ld/129,131,132,133,134,136,137,138,139,140,140,142,144,142,  &
       &       127,124,131,136,154,84/
  data lm/30,0,10,30,20,0,10,30,50,50,15,15,15,0,30,0,0,0,0,40/

  pi=acos(-1.0d0)
  a=6378137.0d0
  f=298.257222101d0
  m0=0.9999d0
  n=1.0d0/(2.0d0*f-1.0d0)
!  
  phi0=ibd(KEI)*pi/180.0d0
  lam0=(ld(KEI)+lm(KEI)/60.0d0)*pi/180.0d0
!
  temp=2.0d0*sqrt(n)/(1.0d0+n)
! temp=temp*atanh(temp*sin(phi))
  temp=temp*0.5d0*log((1.0d0+temp*sin(phi))/(1.0d0-temp*sin(phi)))
! t=sinh(atanh(sin(phi))-temp)
  t=sinh(0.5d0*log((1.0d0+sin(phi))/(1.0d0-sin(phi)))-temp)
  tvar=sqrt(1.0d0+t**2)
!
  lamc=cos(lam-lam0)
  lams=sin(lam-lam0)
  gzd =atan(t/lamc)
!  etad=atanh(lams/tvar)
  etad=0.5d0*log((1.0d0+lams/tvar)/(1.0d0-lams/tvar))
!
  alpha1=   -127.0d0/288.0d0*n**5 +       41.0d0/180.0d0*n**4 +  5.0d0/16.0d0*n**3 &
       &          -   2.0d0/3.0d0*n**2  + 0.5d0*n 
  alpha2=    281.0d0/630.0d0*n**5 +     557.0d0/1440.0d0*n**4 -         0.6d0*n**3 &
       &          + 13.0d0/48.0d0*n**2
  alpha3=15061.0d0/26880.0d0*n**5 -      103.0d0/140.0d0*n**4 + 61.0d0/240.d0*n**3
  alpha4=   -179.0d0/168.0d0*n**5 + 49561.0d0/161280.0d0*n**4 
  alpha5=34729.0d0/80640.0d0*n**5 
!
  a0= 0.015625d0*n**4 + 0.25d0*n**2 + 1.0d0
  a1= ( 0.015625d0*n**5 + 0.125d0*n**3 - n )*1.5d0
  a2= (-0.25d0*n**4+n**2) * 0.9375d0
  a3= ( 0.3125d0*n**5-n**3) * 35.0d0/48.0d0
  a4= 0.615234375d0*n**4
  a5=-693.0d0/1280.0d0*n**5
!
  sphi0var= a5*sin(10.0d0*phi0) + a4*sin( 8.0d0*phi0) + a3*sin( 6.0d0*phi0)  &
       &  + a2*sin( 4.0d0*phi0) + a1*sin( 2.0d0*phi0) + a0*phi0
  sphi0var=sphi0var*m0*a/(1.0d0+n)
!
  avar=m0*a*a0/(1.0d0+n)
!
  temp=      alpha5*sin(10.0d0*gzd)*cosh(10.0d0*etad) + alpha4*sin( 8.0d0*gzd)*cosh( 8.0d0*etad) &
       &   + alpha3*sin( 6.0d0*gzd)*cosh( 6.0d0*etad) + alpha2*sin( 4.0d0*gzd)*cosh( 4.0d0*etad) &
       &   + alpha1*sin( 2.0d0*gzd)*cosh( 2.0d0*etad) + gzd
  xx=temp*avar-sphi0var
!
  temp=      alpha5*cos(10.0d0*gzd)*sinh(10.0d0*etad) + alpha4*cos( 8.0d0*gzd)*sinh( 8.0d0*etad) &
       &   + alpha3*cos( 6.0d0*gzd)*sinh( 6.0d0*etad) + alpha2*cos( 4.0d0*gzd)*sinh( 4.0d0*etad) &
       &   + alpha1*cos( 2.0d0*gzd)*sinh( 2.0d0*etad) + etad
  yy=temp*avar

end subroutine bl2xy
